/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */

#pragma once

#ifdef OX_USE_STDLIB
#include <string_view>
#endif

#include "iterator.hpp"
#include "strops.hpp"
#include "types.hpp"
#include "bit.hpp"

namespace ox {

template<std::size_t buffLen>
class BString;

template<std::size_t buffLen>
class BasicString;

class StringView {
	public:
		template<typename RefType = char&, typename PtrType = char*, bool reverse = false>
		struct iterator: public Iterator<std::bidirectional_iterator_tag, char> {
			private:
				PtrType m_t = nullptr;
				std::size_t m_offset = 0;
				std::size_t m_max = 0;

			public:
				constexpr iterator() noexcept = default;

				constexpr iterator(PtrType t, std::size_t offset, std::size_t max) noexcept {
					m_t = t;
					m_offset = offset;
					m_max = max;
				}

				constexpr auto offset() const noexcept {
					return m_offset;
				}

				constexpr iterator operator+(std::size_t s) const noexcept {
					if constexpr(reverse) {
						return iterator(m_t, max<std::size_t>(m_offset - s, 0), m_max);
					} else {
						return iterator(m_t, min<std::size_t>(m_offset + s, m_max), m_max);
					}
				}

				constexpr auto operator-(const iterator &other) const noexcept {
					if constexpr(reverse) {
						return m_offset + other.m_offset;
					} else {
						return m_offset - other.m_offset;
					}
				}

				constexpr iterator operator-(std::size_t s) const noexcept {
					if constexpr(reverse) {
						return iterator(m_t, min<std::size_t>(m_offset + s, m_max), m_max);
					} else {
						return iterator(m_t, max<std::size_t>(m_offset - s, 0), m_max);
					}
				}

				constexpr iterator &operator+=(std::size_t s) noexcept {
					if constexpr(reverse) {
						m_offset = max<std::size_t>(m_offset - s, 0);
					} else {
						m_offset = min(m_offset + s, m_max);
					}
					return *this;
				}

				constexpr iterator &operator-=(std::size_t s) noexcept {
					if constexpr(reverse) {
						m_offset = min(m_offset + s, m_max);
					} else {
						m_offset = max<std::size_t>(m_offset - s, 0);
					}
					return *this;
				}

				constexpr iterator &operator++() noexcept {
					return operator+=(1);
				}

				constexpr iterator &operator--() noexcept {
					return operator-=(1);
				}

				constexpr RefType operator*() const noexcept {
					return m_t[m_offset];
				}

				constexpr RefType operator[](std::size_t s) const noexcept {
					return m_t[s];
				}

				constexpr bool operator<(const iterator &other) const noexcept {
					return m_offset < other.m_offset;
				}

				constexpr bool operator>(const iterator &other) const noexcept {
					return m_offset > other.m_offset;
				}

				constexpr bool operator<=(const iterator &other) const noexcept {
					return m_offset <= other.m_offset;
				}

				constexpr bool operator>=(const iterator &other) const noexcept {
					return m_offset >= other.m_offset;
				}

				constexpr bool operator==(const iterator &other) const noexcept {
					return m_t == other.m_t && m_offset == other.m_offset && m_max == other.m_max;
				}

				constexpr bool operator!=(const iterator &other) const noexcept {
					return m_t != other.m_t || m_offset != other.m_offset || m_max != other.m_max;
				}

		};

	private:
		const char *m_str = nullptr;
		std::size_t m_len = 0;
	public:
		constexpr StringView() noexcept = default;

		constexpr StringView(const StringView &sv) noexcept = default;

#ifdef OX_USE_STDLIB
		constexpr StringView(const std::string_view &sv) noexcept: m_str(sv.data()), m_len(sv.size()) {}
#endif

		template<std::size_t SmallStrSz>
		constexpr StringView(const BasicString<SmallStrSz> &str) noexcept: m_str(str.c_str()), m_len(str.len()) {}

		template<std::size_t SmallStrSz>
		constexpr StringView(const BString<SmallStrSz> &str) noexcept: m_str(str.c_str()), m_len(str.len()) {}

		constexpr StringView(std::nullptr_t) noexcept {}

		constexpr StringView(const char *str) noexcept: m_str(str), m_len(str ? ox_strlen(str) : 0) {}

		constexpr StringView(const char *str, std::size_t len) noexcept: m_str(str), m_len(len) {}

		[[nodiscard]]
		constexpr iterator<const char&, const char*> begin() const noexcept {
			return {m_str, 0, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*> end() const noexcept {
			return {m_str, m_len, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*> cbegin() const noexcept {
			return {m_str, 0, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*> cend() const noexcept {
			return {m_str, m_len, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*, true> crbegin() const noexcept {
			return {m_str, m_len - 1, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*, true> crend() const noexcept {
			return {m_str, MaxValue<std::size_t>, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*, true> rbegin() const noexcept {
			return {m_str, m_len - 1, m_len};
		}

		[[nodiscard]]
		constexpr iterator<const char&, const char*, true> rend() const noexcept {
			return {m_str, MaxValue<std::size_t>, m_len};
		}

		[[nodiscard]]
		constexpr auto bytes() const noexcept {
			return m_len;
		}

		[[nodiscard]]
		constexpr auto len() const noexcept {
			return m_len;
		}

		[[nodiscard]]
		constexpr auto *data() const noexcept {
			return &m_str[0];
		}

		[[nodiscard]]
		constexpr auto &front() const noexcept {
			return m_str[0];
		}

		[[nodiscard]]
		constexpr auto &back() const noexcept {
			return m_str[m_len - 1];
		}

		constexpr auto substr(std::size_t pos) const noexcept {
			return StringView(m_str + pos, m_len - pos);
		}

		constexpr auto operator[](std::size_t i) const noexcept {
			return m_str[i];
		}

		constexpr auto operator=(const StringView &other) noexcept {
			m_str = other.m_str;
			m_len = other.m_len;
			return *this;
		}

		constexpr auto operator==(const StringView &other) const noexcept {
			if (other.len() != len()) {
				return false;
			}
			return ox_strncmp(m_str, other.m_str, len()) == 0;
		}

		constexpr auto operator<=>(const StringView &str2) const noexcept {
			const StringView &str1 = *this;
			const auto maxLen = ox::min(str1.len(), str2.len());
			const auto a = &str1.front();
			const auto b = &str2.front();
			for (std::size_t i = 0; i < maxLen && (a[i] || b[i]); ++i) {
				if (a[i] < b[i]) {
					return -1;
				} else if (a[i] > b[i]) {
					return 1;
				}
			}
			if (str1.len() > str2.len()) {
				return 1;
			} else if (str1.len() < str2.len()) {
				return -1;
			} else {
				return 0;
			}
		}

};

using CRStringView = const StringView&;

[[nodiscard]]
constexpr bool beginsWith(CRStringView base, CRStringView beginning) noexcept {
	const auto beginningLen = ox::min(beginning.len(), base.len());
	return ox_strncmp(base.data(), beginning, beginningLen) == 0;
}

[[nodiscard]]
constexpr bool endsWith(CRStringView base, CRStringView ending) noexcept {
	const auto endingLen = ending.len();
	return base.len() >= endingLen && ox_strcmp(base.data() + (base.len() - endingLen), ending) == 0;
}

#ifdef OX_USE_STDLIB
constexpr auto toStdStringView(CRStringView sv) noexcept {
	return std::string_view(sv.data(), sv.bytes());
}
#endif

}
